# coding: utf-8

__all__ = ['atomic_number', 'atomic_symbol', 'atomic_mass']

import numpy as np
import pandas as pd

from .atomicdata import atomicdata

class AtomicInfo():
    
    def __init__(self):
        """Class initializer"""
        self.__data = atomicdata
    
    @property
    def data(self):
        return self.__data
    
    @property
    def renames(self):
        """dict : Matches systematic element symbols to their now assigned symbols"""
        return {
            'Unq': 'Rf',
            'Unp': 'Db',
            'Unh': 'Sg',
            'Uns': 'Bh',
            'Uno': 'Hs',
            'Une': 'Mt',
            'Uun': 'Ds',
            'Uuu': 'Rg',
            'Uub': 'Cn',
            'Uut': 'Nh',
            'Uuq': 'Fl',
            'Uup': 'Mc',
            'Uuh': 'Lv',
            'Uus': 'Ts',
            'Uuo': 'Og',
        }        
    
    def load(self, datafile):
        """
        Reads in the "Linearized ASCII Output" as found at https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses
        """
        with open(datafile) as f:
            lines = f.readlines()

        data = []
        i = 0
        while i < len(lines):
            if len(lines[i]) > 15 and lines[i][:15] == 'Atomic Number =':
                d = {}
                while i < len(lines) and len(lines[i].strip()) > 0:
                    terms = lines[i].split('=')
                    if len(terms) > 1:
                        d[terms[0].strip()] = terms[1].strip()
                    else:
                        d[terms[0].strip()] = np.nan 
                    i += 1
                data.append(d)
            else:
                i += 1
        data = pd.DataFrame(data)
        
        def make_int(series, key):
            return int(series[key])
        
        data['Atomic Number'] = data.apply(make_int, args=['Atomic Number'], axis=1)
        data['Mass Number'] = data.apply(make_int, args=['Mass Number'], axis=1)
        
        self.__data = data
        
    def atomic_number(self, atomic_symbol):
        """
        Return the corresponding atomic number for a given atomic symbol.

        Parameters
        ----------
        atomic_symbol : str
            An atomic symbol.

        Return
        ------
        int
            The corresponding atomic number.
        """
        
        # Handle old systematic named symbols
        if atomic_symbol in self.renames:
            atomic_symbol = self.renames[atomic_symbol]
        
        matches = self.data[self.data['Atomic Symbol'] == atomic_symbol]
        if len(matches) > 0:
            return matches.iloc[0]['Atomic Number']
        else:
            raise ValueError(f'No matches for atomic symbol {atomic_symbol} found')
    
    def atomic_symbol(self, atomic_number):
        """
        Return the corresponding atomic symbol for a given atomic number.

        Parameters
        ----------
        atomic_number : int
            An atomic number.

        Return
        ------
        int
            The corresponding atomic symbol.
        """
        matches = self.data[self.data['Atomic Number'] == atomic_number]
        if len(matches) > 0:
            return matches.iloc[0]['Atomic Symbol']
        else:
            raise ValueError(f'No matches for atomic number {atomic_number} found')
    
    def atomic_mass(self, atomic_info, mass_number=None, prompt=True):
        """
        Returns either the median standard atomic weight for an element or the relative
        atomic mass for an isotope.
        
        Parameters
        ----------
        atomic_info : str or int
            The atomic symbol or number identifying the element/isotopes.  
        mass_number : int, optional
            An isotope mass number.
        Returns
        -------
        float
            The average standard atomic weight of an element or the relative
            atomic mass of an isotope.
        """

        # Try converting atomic_info to an int - fetch atomic_symbol if needed
        try:
            atomic_number = int(atomic_info)
        except:
            atomic_symbol = atomic_info
        else:
            atomic_symbol = self.atomic_symbol(atomic_number)        
            
        # Separate mass_number from atomic_symbol if needed
        if '-' in atomic_symbol:
            if mass_number is not None:
                raise ValueError('Mass number cannot be given both with the symbol and separately')
            atomic_symbol, mass_number = atomic_symbol.split('-')
        
        # Handle hydrogens
        if atomic_symbol in ['H', 'D', 'T']:
            atomic_symbol, mass_number = self.__handle_hydrogen(atomic_symbol, mass_number=mass_number)
        
        # Handle old systematic named symbols
        if atomic_symbol in self.renames:
            atomic_symbol = self.renames[atomic_symbol]
        
        # Check if there is a standard atomic weight for an element
        if mass_number is None:
            matches = self.data[self.data['Atomic Symbol'] == atomic_symbol]
            weight = matches.iloc[0]['Standard Atomic Weight']
            if isinstance(weight, float):
                if pd.notna(weight):
                    return weight
            elif '[' in weight:
                return np.array(eval(weight)).mean()
            elif '(' in weight:
                return float(weight.split('(')[0])
            elif weight != '':
                return float(weight)
            elif len(matches) == 0:
                raise ValueError(f'No matches for atomic symbol {atomic_symbol} found')
        
            # Return isotope mass if only one isotope
            if len(matches) == 1:
                return float(matches.iloc[0]['Relative Atomic Mass'].split('(')[0])
            
            if prompt:
                print(f'No standard atomic weight for {atomic_symbol}.')
                isotopes = matches['Mass Number'].to_list()
                print(f'Please select an isotope from {isotopes}:')
                mass_number = input()
            else:
                raise ValueError(f'No standard weight for {atomic_symbol}. Specify an isotope instead.')
        
        # Convert mass_number to an int
        mass_number = int(mass_number)
                      
        # Find relative atomic mass
        matches = self.data[(self.data['Atomic Symbol'] == atomic_symbol) 
                           &(self.data['Mass Number'] == mass_number)]
        
        if len(matches) == 1:
            mass = matches.iloc[0]['Relative Atomic Mass']
            if isinstance(mass, float):
                return mass
            elif '(' in mass:
                return float(mass.split('(')[0])
            else:
                raise ValueError('Mass value format not recognized!!!!!!!!!!')

        elif len(matches) == 0:
            raise ValueError(f'No matches for atomic symbol {atomic_symbol} and mass number {mass_number} found')
        else:
            raise ValueError('Multiple matches found!!!')
        
    def __handle_hydrogen(self, atomic_symbol, mass_number=None):
        """
        Special handling of hydrogen isotopes due to isotope symbols D and T
        
        Parameters
        ----------
        atomic_symbol : str
            The atomic symbol.  
        mass_number : int, optional
            An isotope mass number.
        
        Returns
        -------
        atomic_symbol : str
            The correct database atomic symbol for the input
        mass_number : int or None
            The correct database mass number for the input
        """
        if mass_number is not None:
            mass_number = int(mass_number)
            
        # Check that mass number is 2 for 'D'
        if atomic_symbol == 'D':
            if mass_number is None:
                mass_number = 2
            elif mass_number != 2:
                raise ValueError('Mass number must be 2 for isotope symbol D')
        
        # Check that mass number is 3 for 'T'
        elif atomic_symbol == 'T':
            if mass_number is None:
                mass_number = 3
            elif mass_number != 3:
                raise ValueError('Mass number must be 3 for isotope symbol T')
        else:
            # Change 'H' to 'D' for mass number 2
            if mass_number == 2:
                atomic_symbol = 'D'
            # Change 'H' to 'T' for mass number 3
            elif mass_number == 3:
                atomic_symbol = 'T'
                
        return atomic_symbol, mass_number

# Initialize an object of the class and create shortcuts to the methods
atomicinfo = AtomicInfo()
atomic_number = atomicinfo.atomic_number
atomic_symbol = atomicinfo.atomic_symbol
atomic_mass = atomicinfo.atomic_mass


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# atomman imports
import atomman.unitconvert as uc
from .process_prop_info import process_prop_info
from ... import Atoms, System
from ...tools import uber_open_rmode

def load(table, box, symbols=None, system=None, prop_name=None, table_name=None,
         shape=None, unit=None, dtype=None, prop_info=None, skiprows=None,
         nrows=None, comment=None, usecols=None, header='infer'):
    """
    Reads in tabular data into atomic properties.
    
    Parameters
    ----------
    table : str or file-like object
        The table content, file path or file-like object containing the
        content to read.
    box : atomman.Box
        The atomic box to use when generating a System around the data.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
    system : atomman.System, optional
        The atomic system to load the values to.  If not given, a new system
        will be constructed.
    prop_name : list, optional
         The Atoms properties to generate.  Must be given if prop_info is not.
    table_name : list, optional
        The table column name(s) that correspond to each prop_name.  If not
        given, the table_name values will be based on the prop_name values.
    shape : list, optional
        The shape of each per-atom property.  If not given, will be inferred
        from the length of each table_name value.
    unit : list, optional
        Lists the units for each prop_name as stored in the table.  For a
        value of None, no conversion will be performed for that property.  For
        a value of 'scaled', the corresponding table values will be taken in
        box-scaled units.  If not given, all unit values will be set to None
        (i.e. no conversions).
    dtype : list, optional
        Allows for the data type of each property to be explicitly given.
        Values of None will infer the data type from the corresponding
        property values.  If not given, all values will be None.
    prop_info : list of dict, optional
        Structured form of property conversion parameters, in which each
        dictionary in the list corresponds to a single atoms property.  Each
        dictionary must have a 'prop_name' field, and can optionally have
        'table_name', 'shape', 'unit', and 'dtype' fields.
    skiprows : int, optional
        Number of rows to skip before reading the data.
    usecols : int, optional
        Which columns are to be used. Will be passed to pandas.read_csv() usecols
        option.
    nrows : int, optional
        Number of rows of data to read.
    comment : str, optional
        Specifies a character which indicates all text on a given line after
        is to be considered to be a comment and ignored by parser. This is
        often '#'.
        
    Returns
    -------
    atomman.System
        The generated system.
    """
    # Process conversion parameters
    prop_info = process_prop_info(prop_name=prop_name, table_name=table_name,
                                  shape=shape, unit=unit, dtype=dtype, prop_info=prop_info)
    
    # Build list of all table_names
    table_name = []
    for prop in prop_info:
        table_name += prop['table_name']
    
    # Read in table to dataframe
    with uber_open_rmode(table) as f:
        df = pd.read_csv(f, delim_whitespace=True, names=table_name, skiprows=skiprows,
                         nrows=nrows, comment=comment, header=header, usecols=usecols)
    if 'id' in df:
        df = df.sort_values('id')
    
    # Generate System
    natoms = len(df)
    if system is None:
        system = System(atoms=Atoms(natoms=natoms), box=box)
    
    # Loop over all properties
    for prop in prop_info:
        pname = prop['prop_name']
        if pname == 'a_id':
            continue
        
        # Get values
        value = df[prop['table_name']].values.reshape((natoms,) + prop['shape'])
        
        if prop['unit'] is not None:
            if prop['unit'] == "scaled":
                value = system.unscale(value)
            else:
                value = uc.set_in_units(value, prop['unit'])
        value = np.asarray(value, dtype=prop['dtype'])
        system.atoms.view[pname] = value
    
    # Set symbols if given
    if symbols is not None:
        system.symbols = symbols
    
    return system


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from ... import Atoms, Box, System
from ...tools import uber_open_rmode

def load(poscar, symbols=None, prop={}):
    """
    Reads a poscar-style coordination file for a system.
    
    Parameters
    ----------
    poscar : str or file-like object
        The POSCAR content to read
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
        Useful if the symbols for the model differ from the standard element
        tags or if the poscar file has no elemental information.
    prop : dict, optional
        Dictionary containing any extra per-atom properties to include.
    
    Returns
    -------
    system : atomman.System
        The system object associated with the poscar file.
    """
    
    # Read in all lines of the file
    with uber_open_rmode(poscar) as f:
        lines = f.readlines()
    
    # Interpret box information
    box_scale = float(lines[1])
    avect = np.array(lines[2].split(), dtype='float64') * box_scale
    bvect = np.array(lines[3].split(), dtype='float64') * box_scale
    cvect = np.array(lines[4].split(), dtype='float64') * box_scale
    box = Box(avect=avect, bvect=bvect, cvect=cvect)
    
    # Read in elements, number of types, and style info
    try:
        typenums = np.array(lines[5].split(), dtype='int64')
        elements = [None for n in range(len(typenums))]
        style = lines[6]
        start_i = 7
    except:
        elements = lines[5].split()
        typenums = np.array(lines[6].split(), dtype='int64')
        style = lines[7]
        start_i = 8
    
    # Decode elements if needed
    for i in range(len(elements)):
        try:
            elements[i] = elements[i].decode('UTF-8')
        except:
            pass
    
    # Handle elements and symbols
    if symbols is None:
        symbols = elements
    
    # Build atype list
    atype = np.array([], dtype='int64')
    for i in range(len(typenums)):
        atype = np.hstack((atype, np.full(typenums[i], i+1, dtype='int64')))
    
    # Check which coordinate style to use
    try:
        style = style.decode('UTF-8')
    except:
        pass
    if style[0] in 'cCkK':
        scale = False
    else:
        scale = True
    
    # Read in positions
    natoms = np.sum(typenums)
    pos = np.empty((natoms, 3), dtype='float64')
    count = 0
    for i in range(natoms):
        terms = lines[i+start_i].split()
        if len(terms) > 0:
            pos[count,:] = np.array(terms[:3], dtype='float64')
        count += 1
    
    # Add atype, pos to prop dictionary
    prop['atype'] = atype
    prop['pos'] = pos
    
    atoms = Atoms(prop=prop)
    system = System(atoms=atoms, box=box, scale=scale, symbols=symbols)
    
    return system


# coding: utf-8
# Standard Python libraries
import os
import glob
import shutil
import subprocess as sp

# atomman imports
from . import Log, LammpsError

def run(lammps_command, script_name, mpi_command=None,
        restart_script_name=None, logfile='log.lammps'):
    """
    Calls LAMMPS to run. Returns data model containing LAMMPS output.
    
    Parameters
    ----------
    lammps_command : str
        The LAMMPS inline run command (sans -in script_name).
    script_name : str
        Path of the LAMMPS input script to use.
    mpi_command : str or None, optional
        The MPI inline command to run LAMMPS in parallel. Default value is 
        None (no mpi).
    restart_script_name : str or None, optional
        Alternative script to use for restarting if logfile already exists.
        Default value is None (no restarting).
    logfile : str, optional
        Specifies the path to the logfile to write to.  Default value is
        'log.lammps'.
    
    Returns
    -------
    atomman.lammps.Log or DataModelDict
        The content either as a Log object or in data model format.
    """
    
    # Check if restart_script_name is given
    if restart_script_name is not None:
        
        # Check if simulation was previously started by looking for log.lammps
        if os.path.isfile(logfile):
            logname, logext = os.path.splitext(logfile)
            # Replace script_name with restart_script_name
            script_name = restart_script_name
            
            # Search for any earlier log files with the name log-*.lammps
            logids = []
            for oldlog in glob.iglob(logname + '-*' + logext):
                logids.append(int(os.path.splitext(os.path.basename(oldlog))[0][len(logname)+1:]))
            
            # Rename old logfile to keep it from being overwritten
            if len(logids) == 0:
                lognum = 1
            else:
                lognum = max(logids)+1
            shutil.move(logfile, logname + '-' + str(lognum) + logext)
        else:
            lognum = 0
    else:
        lognum = 0
    
    # Convert lammps_command into list of terms
    if isinstance(lammps_command, str):
        lammps_command = lammps_command.split(' ')
    elif not isinstance(lammps_command, list):
        lammps_command = [lammps_command]
    
    # Convert script_name into list of terms
    if isinstance(script_name, str):
        script_name = script_name.split(' ')
    elif not isinstance(script_name, list):
        script_name = [script_name]
    
    # Convert mpi_command into list of terms
    if mpi_command is None:
        mpi_command = []
    elif isinstance(mpi_command, str):
        mpi_command = mpi_command.split(' ')
    elif not isinstance(mpi_command, list):
        mpi_command = [mpi_command]
    
    # Extra terms
    extra = []
    
    # Check if logfile is not log.lammps
    if logfile != 'log.lammps':
        # Convert logfile into list of terms
        extra += ['-log'] + logfile.split(' ')
    
    # Try to run lammps as a subprocess
    try:
        output = sp.check_output(mpi_command + lammps_command + extra + ['-in'] + script_name)
    
    # Convert LAMMPS error to a Python error if failed
    except sp.CalledProcessError as e:
        raise LammpsError(e.output.decode('UTF-8'))
    
    # Initialize Log object
    log = Log()
    
    # Read in from all old log files
    for i in range(1, lognum+1):
        log.read(logname + '-' + str(i) + logext)
    
    # Read in from current logfile
    log.read(logfile)
    
    return log


# coding: utf-8
# Standard Python libraries
from io import open
from collections import OrderedDict

# atomman imports
import atomman.unitconvert as uc
from ...lammps import style
from .process_prop_info import process_prop_info
from ...tools import indexstr

def dump(system, f=None, lammps_units='metal', scale=False, prop_name=None,
         table_name=None, shape=None, unit=None, dtype=None,
         prop_info=None, float_format ='%.13f', return_prop_info=False):
    """
    Write a LAMMPS-style atom data file from a System.
    
    Parameters
    ----------
    system : atomman.System 
        The system to write to the atom data file.
    f : str or file-like object, optional
        File path or file-like object to write the content to.  If not given,
        then the content is returned as a str.
    lammps_units : str, optional
        The LAMMPS units option associated with the table values.  This is used
        for the box dimensions and default units for standard dump properties 
        (not compute/fix definitions). 
    scale : bool, optional
        Flag indicating if atom positions are to be scaled relative to the box
        (True) or given in absolute Cartesian values (False, default).
    prop_name : list, optional
        The Atoms properties to include.  If neither prop_name or prop_info are
        given, all system properties will be included.
    table_name : list, optional
        The dump table column name(s) that correspond to each prop_name.  If not
        given, the table_name values will be based on the prop_name and shape
        values.
    shape : list, optional
        The shape of each per-atom property.  If not given, will be inferred
        from the length of each table_name value.
    unit : list, optional
        Lists the units for each prop_name as stored in the table.  For a
        value of None, no conversion will be performed for that property.  For
        a value of 'scaled', the corresponding table values will be taken in
        box-scaled units.  If not given, unit values will be taken based on
        lammps_units if prop_name corresponds to a standard LAMMPS property, 
        otherwise will be set to None (no conversion).
    dtype : list, optional
        Allows for the data type of each property to be explicitly given.
        Values of None will infer the data type from the corresponding
        property values.  If not given, all values will be None.
    prop_info : list of dict, optional
        Structured form of property conversion parameters, in which each
        dictionary in the list corresponds to a single atoms property.  Each
        dictionary must have a 'prop_name' field, and can optionally have
        'table_name', 'shape', 'unit', and 'dtype' fields.
    float_format : str, optional
        c-style formatting string for floating point values.  Default value is
        '%.13f'.
    return_prop_info : bool, optional
        Flag indicating if the filled-in prop_info is to be returned.  Having
        this allows for 1:1 load/dump conversions.  Default value is False
        (prop_info is not returned).
        
    Returns
    -------
    content : str
        The generated atom_data content.  Only returned if f is None.
    
    prop_info : list of dict
        The filled-in prop_info structure. Only returned if
        return_prop_info is True.
    """
    lammps_unit = style.unit(lammps_units)
    
    # Set default values
    if prop_info is None:
        if prop_name is None:
            atoms_props = system.atoms_prop()
            try:
                atoms_props.pop(atoms_props.index('atom_id'))
            except:
                pass
            prop_name = ['atom_id'] + atoms_props
        
        if shape is None and table_name is None:
            shape = []
            for name in prop_name:
                if name == 'atom_id':
                    shape.append(())
                elif name in ['spos', 'upos', 'supos']:
                    shape.append((3,))
                else:
                    shape.append(system.atoms.view[name].shape[1:])
    
    # Process conversion parameters
    prop_info = process_prop_info(prop_name=prop_name, table_name=table_name,
                                  shape=shape, unit=unit, dtype=dtype,
                                  prop_info=prop_info,
                                  lammps_units=lammps_units)
    
    # Write timestep info
    content = 'ITEM: TIMESTEP\n'
    try:
        content += '%i\n' % system.timestep
    except:
        content += '0\n'
    
    # Write number of atoms
    content += 'ITEM: NUMBER OF ATOMS\n'
    content += '%i\n' % (system.natoms)
    
    # Extract and convert box values
    xlo = uc.get_in_units(system.box.xlo, lammps_unit['length'])
    xhi = uc.get_in_units(system.box.xhi, lammps_unit['length'])
    ylo = uc.get_in_units(system.box.ylo, lammps_unit['length'])
    yhi = uc.get_in_units(system.box.yhi, lammps_unit['length'])
    zlo = uc.get_in_units(system.box.zlo, lammps_unit['length'])
    zhi = uc.get_in_units(system.box.zhi, lammps_unit['length'])
    xy = uc.get_in_units(system.box.xy, lammps_unit['length'])
    xz = uc.get_in_units(system.box.xz, lammps_unit['length'])
    yz = uc.get_in_units(system.box.yz, lammps_unit['length'])
    
    # Compute absolute box bounds
    xlo_bound = xlo + min((0.0, xy, xz, xy + xz))
    xhi_bound = xhi + max((0.0, xy, xz, xy + xz))
    ylo_bound = ylo + min((0.0, yz))
    yhi_bound = yhi + max((0.0, yz))
    zlo_bound = zlo
    zhi_bound = zhi
    
    is_orthogonal = (xy == 0.0 and xz == 0.0 and yz == 0.0)
    
    # Write system boundary info
    if is_orthogonal:
        content += 'ITEM: BOX BOUNDS'
    else:
        content += 'ITEM: BOX BOUNDS xy xz yz'
        
    # Write pbc info
    for i in range(3):
        if system.pbc[i]:
            content += ' pp'
        else:
            content += ' fm'
    content += '\n'
        
    # Write system boundary info
    if is_orthogonal:
        xf2 = float_format + ' ' + float_format + '\n'
        content += xf2 % (xlo_bound, xhi_bound)
        content += xf2 % (ylo_bound, yhi_bound)
        content += xf2 % (zlo_bound, zhi_bound)
    else:
        xf3 = float_format + ' ' + float_format + ' ' + float_format + '\n'
        content += xf3 % (xlo_bound, xhi_bound, xy)
        content += xf3 % (ylo_bound, yhi_bound, xz)
        content += xf3 % (zlo_bound, zhi_bound, yz)
    
    # Write atomic header info and prepare outarray for writing
    header = 'ITEM: ATOMS'
    for prop in prop_info:
        header += ' ' + ' '.join(prop['table_name'])
    header += '\n'
    content += header
    
    content += table_dump(system, prop_info=prop_info, float_format=float_format)
    
    returns = []
    
    # Save to the file-like object
    if hasattr(f, 'write'):
        f.write(content)
    
    # Save to the file name
    elif f is not None:
        with open(f, 'w') as fp:
            fp.write(content)
    
    # Return as a string
    else:
        returns.append(content)
    
    if return_prop_info is True:
        returns.append(prop_info)
        
    if len(returns) == 1:
        return returns[0]
    elif len(returns) > 1:
        return tuple(returns)
        
def table_dump(system, f=None, prop_info=None, float_format ='%.13f'):
    """
    Converts a system's atoms' values to a string table.  Modified from
    table.dump to handle alternate pos fields that dump files can use.
    
    Parameters
    ----------
    system : atomman.System
        An atomman representation of a system.
    f : str or file-like object, optional
        File path or file-like object to write the content to.  If not given,
        then the content is returned as a str.
    prop_info : list of dict, optional
        Structured form of property conversion parameters, in which each
        dictionary in the list corresponds to a single atoms property.  Each
        dictionary must have a 'prop_name' field, and can optionally have
        'table_name', 'shape', 'unit', and 'dtype' fields.
    float_format : str, optional
        c-style formatting string for floating point values.  Default value is
        '%.13f'.
    
    Returns
    -------
    str
        The generated data table.  Only returned if fp is None.
    """
    # Set parameters
    natoms = system.natoms
    key_rename = OrderedDict()
    
    # Build list of properties to scale and alternate pos
    scale = []
    altpos = []
    for prop in prop_info:
        if prop['prop_name'] in ['spos', 'upos', 'supos']:
            altpos.append(prop['prop_name'])
        if prop['unit'] == 'scaled':
            scale.append(prop['prop_name'])
            prop['unit'] = None
    
    # Transform to dataframe
    df = system.atoms_df(scale)
    
    # Check atom_id values
    if 'atom_id' not in df:
        df['atom_id'] = range(1, natoms+1)
    assert len(df.atom_id) == len(set(df.atom_id)), 'atom_id is not unique for all atoms'    
    
    # Add alternate pos terms
    if 'upos' in altpos:
        df['upos[0]'] = df['pos[0]']
        df['upos[1]'] = df['pos[1]']
        df['upos[2]'] = df['pos[2]']
    if 'spos' in altpos or 'supos' in altpos:
        spos = system.atoms_prop(key='pos', scale=True)
        if 'spos' in altpos:
            df['spos[0]'] = spos[:,0]
            df['spos[1]'] = spos[:,1]
            df['spos[2]'] = spos[:,2]
        if 'supos' in altpos:
            df['supos[0]'] = spos[:,0]
            df['supos[1]'] = spos[:,1]
            df['supos[2]'] = spos[:,2]
    
    # Loop over all properties
    for prop in prop_info:
        pname = prop['prop_name']
        
        # loop over all table names and property indexes
        for tname, (index, istr) in zip(prop['table_name'], # pylint: disable=unused-variable
                                        indexstr(prop['shape'])):
            
            # Build name change dict
            key_rename[pname + istr] = tname
            
            # Convert units if needed
            if prop['unit'] is not None:
                df[pname + istr] = uc.get_in_units(df[pname + istr], prop['unit'])
    
    # Rename and reorganize
    df = df.rename(columns=key_rename)[list(key_rename.values())]
  
    # Generate table
    sep = ' '
    return df.to_csv(path_or_buf=f, sep=sep, index=None, header=False,
                     float_format=float_format, line_terminator='\n')


# coding: utf-8

# http://www.numpy.org/
import numpy as np

def central_difference(fxn, coord, shift=1e-5):
    """
    Computes the gradient of a function at a set of coordinates.
    
    Parameters
    ----------
    fxn : function
        The function to compute gradients for.
    coord : array-like object
        The coordinates to evaluate the gradient at. The number of
        derivatives calculated will be based on the size of the
        coordinate's final dimension, i.e. if coord is (10, 3), then three
        derivatives will be computed for each of the ten coordinates.
    shift : float
        The shift step size to use when evaluating the derivatives.
    
    Returns
    -------
    gradient : array-like object
        The gradient array with the same shape as coord.
    """
    coord = np.asarray(coord)
    
    # Identify number of derivative dimensions based on final array dimension
    ndim = coord.shape[-1]
    
    # Build gradient array
    gradient = np.zeros_like(coord)
    fps = []
    fms = []
    for i in range(ndim):
        
        # Create δ vector to displace along one dimension
        δ = np.zeros(ndim)
        δ[i] = shift
        
        # Compute gradients for that dimension
        fplus = fxn(coord + δ)
        fminus = fxn(coord - δ)
        fps.append(fplus)
        fms.append(fminus)
        
        gradient[..., i] = (fplus - fminus) / (2 * shift)
    
    return gradient


def iaslist(term):
    """
    Iterate over items in term as if term was a list. Treats a str, unicode
    term as a single item.
    
    Parameters
    ----------
    term : any
        Term to iterate over.
    
    Yields
    ------
    any
        Items in the list representation of term.
    """
    if isinstance(term, str):
        yield term
    else:
        try:
            for t in term:
                yield t
        except:
            yield term
            
def aslist(term):
    """
    Create list representation of term. Treats a str, unicode term as a single
    item.
    
    Parameters
    ----------
    term : any
        Term to convert into a list, if needed.
        
    Returns
    -------
    list of any
        All items in term as a list
    """
    return [t for t in iaslist(term)]


# coding: utf-8
# atomman imports
from atomman.lammps import style

def velocities_prop_info(atom_style='atomic', units='metal'):
    """
    Returns the prop_info associated with a Velocity section in a LAMMPS data file
    
    Parameters
    ----------
    atom_style : str, optional
        The LAMMPS atom_style option.  Default value is 'atomic'.
    units : str, optional
        The LAMMPS units option.  Default value is 'metal'.
        
    Returns
    -------
    prop_info : list of dict
        The proper conversion metadata as used by atomman.convert.atoms_table
    """
    lammps_unit = style.unit(units)
    
    if atom_style in ('angle', 'atomic', 'body', 'bond', 'charge', 'dipole',
                      'full', 'line', 'meso', 'molecular', 'peri', 'smd',
                      'template', 'tri', 'wavepacket'):
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "velocity",
                      "table_name": ["vx", "vy", "vz"],
                      "unit": lammps_unit['velocity']}]
    
    elif atom_style == 'electron':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "velocity",
                      "table_name": ["vx", "vy", "vz"],
                      "unit": lammps_unit['velocity']},
                     
                     {"prop_name": "eradial_velocity",
                      "table_name": "ervel",
                      "unit": lammps_unit['velocity']}]
    
    elif atom_style == 'ellipsoid':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "velocity",
                      "table_name": ["vx", "vy", "vz"],
                      "unit": lammps_unit['velocity']},
                     
                     {"prop_name": "ang_momentum",
                      "table_name": ["lx", "ly", "lz"],
                      "unit": lammps_unit['ang-mom']}]
    
    elif atom_style == 'sphere':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "velocity",
                      "table_name": ["vx", "vy", "vz"],
                      "unit": lammps_unit['velocity']},
                     
                     {"prop_name": "ang_velocity",
                      "table_name": ["wx", "wy", "wz"],
                      "unit": lammps_unit['ang-vel']}]
    
    elif atom_style[:6] == 'hybrid':
        substyles = atom_style.split()
        prop_info = velocities_prop_info('atomic')
        for substyle in substyles[1:]:
            prop_names = []
            for prop in prop_info:
                prop_names.append(prop['prop_name'])
            subprop_info = velocities_prop_info(substyle)
            for prop in subprop_info:
                if prop['prop_name'] not in prop_names:
                    prop_info.append(prop)
    else:
        raise ValueError('atom_style ' + atom_style + ' not supported')
    
    return prop_info


# coding: utf-8
# Standard Python libraries
from collections import OrderedDict

def unit(units='metal'):
    """
    Returns the units information associated with a LAMMPS units option.
    
    Parameters
    ----------
    units : str, optional
        The LAMMPS unit option.  Default value is 'metal'
    
    Returns
    -------
    dict
        The keys are physical quantity types and the values are the associated
        units to use.
    """
    params = OrderedDict()
    
    if units == 'lj':
        params['mass'] =                None
        params['length'] =              None
        params['time'] =                None
        params['energy'] =              None
        params['velocity'] =            None
        params['force'] =               None
        params['torque'] =              None
        params['temperature'] =         None
        params['pressure'] =            None
        params['dynamic viscosity'] =   None
        params['charge'] =              None
        params['dipole'] =              None
        params['electric field'] =      None
        params['density'] =             None
    
    elif units == 'real':
        params['mass'] =                'g/mol'
        params['length'] =              'angstrom'
        params['time'] =                'fs'
        params['energy'] =              'kcal/mol'
        params['velocity'] =            'angstrom/fs'
        params['force'] =               'kcal/(mol*angstrom)'
        params['torque'] =              'kcal/mol'
        params['temperature'] =         'K'
        params['pressure'] =            'atm'
        params['dynamic viscosity'] =   'Pa*s/10'
        params['charge'] =              'e'
        params['dipole'] =              'e*angstrom'
        params['electric field'] =      'V/angstrom'
        params['density'] =             'g/cm^3'
    
    elif units == 'metal':
        params['mass'] =                'g/mol'
        params['length'] =              'angstrom'
        params['time'] =                'ps'
        params['energy'] =              'eV'
        params['velocity'] =            'angstrom/ps'
        params['force'] =               'eV/angstrom'
        params['torque'] =              'eV'
        params['temperature'] =         'K'
        params['pressure'] =            'bar'
        params['dynamic viscosity'] =   'Pa*s/10'
        params['charge'] =              'e'
        params['dipole'] =              'e*angstrom'
        params['electric field'] =      'V/angstrom'
        params['density'] =             'g/cm^3'
    
    elif units == 'si':
        params['mass'] =                'kg'
        params['length'] =              'm'
        params['time'] =                's'
        params['energy'] =              'J'
        params['velocity'] =            'm/s'
        params['force'] =               'N'
        params['torque'] =              'N*m'
        params['temperature'] =         'K'
        params['pressure'] =            'Pa'
        params['dynamic viscosity'] =   'Pa*s'
        params['charge'] =              'C'
        params['dipole'] =              'C*m'
        params['electric field'] =      'V/m'
        params['density'] =             'kg/m^3'
    
    elif units == 'cgs':
        params['mass'] =                'g'
        params['length'] =              'cm'
        params['time'] =                's'
        params['energy'] =              'erg'
        params['velocity'] =            'cm/s'
        params['force'] =               'dyn'
        params['torque'] =              'dyn*cm'
        params['temperature'] =         'K'
        params['pressure'] =            '0.1*Pa'
        params['dynamic viscosity'] =   'Pa*s/10'
        params['charge'] =              '10*c0*C'
        params['dipole'] =              '10*c0*C*cm'
        params['electric field'] =      'c0*uV/cm'
        params['density'] =             'g/cm^3'
    
    elif units == 'electron':
        params['mass'] =                'amu'
        params['length'] =              'aBohr'
        params['time'] =                'fs'
        params['energy'] =              '2*Ry'
        params['velocity'] =            '2*Ry*aBohr/hbar'
        params['force'] =               '2*Ry/aBohr'
        params['torque'] =              'dyn*cm'
        params['temperature'] =         'K'
        params['pressure'] =            'Pa'
        params['charge'] =              'e'
        params['dipole'] =              '1e-21/c0*C*m'
        params['electric field'] =      'V/cm'
    
    elif units == 'micro':
        params['mass'] =                'pg'
        params['length'] =              'um'
        params['time'] =                'us'
        params['energy'] =              'pg*um^2/us^2'
        params['velocity'] =            'um/us'
        params['force'] =               'pg*um/us^2'
        params['torque'] =              'pg*um^2/us^2'
        params['temperature'] =         'K'
        params['pressure'] =            'pg/(um*us^2)'
        params['dynamic viscosity'] =   'pg/(um*us)'
        params['charge'] =              '1e-12*C'
        params['dipole'] =              '1e-12*C*um'
        params['electric field'] =      'V/um'
        params['density'] =             'pg/um^3'
    
    elif units == 'nano':
        params['mass'] =                '1e-18*g'
        params['length'] =              'nm'
        params['time'] =                'ns'
        params['energy'] =              '1e-18*g*nm^2/ns^2'
        params['velocity'] =            'nm/ns'
        params['force'] =               '1e-18*g*nm/ns^2'
        params['torque'] =              '1e-18*g*nm^2/ns^2'
        params['temperature'] =         'K'
        params['pressure'] =            '1e-18*g/(nm*ns^2)'
        params['dynamic viscosity'] =   '1e-18*g/(nm*ns)'
        params['charge'] =              'e'
        params['dipole'] =              'e*nm'
        params['electric field'] =      'V/nm'
        params['density'] =             '1e-18*g/nm^3'
    
    else:
        raise ValueError('units ' + units + ' not supported')
    
    params['ang-mom'] = params['length'] + '*' + params['velocity'] + '*' + params['mass']
    params['ang-vel'] = '1/' + params['time']
    
    return params

def timestep(units='metal'):
    """
    Returns the default timestep value for a LAMMPS unit style option.
    
    Parameters
    ----------
    units : str, optional
        The LAMMPS unit option.  Default value is 'metal'
    
    Returns
    -------
    float
        The default timestep value.
    """
    if units == 'lj':
        return 0.005
    elif units == 'real':
        return 1.0
    elif units == 'metal':
        return 0.001
    elif units == 'si':
        return 1.0e-8
    elif units == 'cgs':
        return 1.0e-8
    elif units == 'electron':
        return 0.001
    elif units == 'micro':
        return 2.0
    elif units == 'nano':
        return 0.00045
    else:
        raise ValueError('units ' + units + ' not supported')


# coding: utf-8

class LammpsError(Exception):
    """
    Class for passing LAMMPS errors to Python
    """
    def __init__(self, *args, **kwargs):
        
        message = args[0]
        try:
            message = message[message.index('ERROR:') + 7:]
        except:
            pass
        args = (message, ) + args[1:]

        super().__init__(*args, **kwargs)

from . import style
from .Log import Log
from .NEBLog import NEBLog
from .Potential import Potential
from .run import run
from .normalize import normalize
from .checkversion import checkversion
__all__ = ['LammpsError', 'style', 'run', 'normalize', 'Potential', 'Log',
           'NEBLog', 'checkversion']


# coding: utf-8

# http://www.numpy.org/
import numpy as np

def dump(system, f=None, header='', symbols=None, coordstyle='direct',
         box_scale=1.0, float_format='%.13e'):
    """
    Generates a poscar-style coordination file for the system.
    
    Parameters
    ----------
    system : atomman.System
        The system whose coordinates you are saving
    f : str or file-like object, optional
        File path or file-like object to write the content to.  If not given,
        then the content is returned as a str.
    header : str, optional
        The comment line to place at the top of the file. Default value is ''.
    symbols : tuple, optional
        List of the element symbols that correspond to the atom types.  If not
        given, will use system.symbols if set, otherwise no element content
        will be included.
    coordstyle : str, optional 
        The poscar coordinate style to use: 'cartesian' or 'direct' (i.e.
        box relative).  Default value is 'direct'.  
    box_scale : float, optional
        A universal scaling constant applied to the box vectors. Default value
        is 1.0.
     float_format : str, optional
        c-style format for printing the floating point numbers. Default value
        is '%.13e'.
        
    Returns
    -------
    poscar_str : str
        String of the poscar object (only returned if fname is not given).
    """
    assert '\n' not in header, 'header can only be one line'
    assert '\n' not in coordstyle, 'coordstyle can only be one line'
    
    threexf = float_format + ' ' + float_format + ' ' + float_format
    
    # Scale box vectors and write out the values
    vects = system.box.vects / box_scale
    poscar_string = '\n'.join([header,
                               float_format % box_scale,
                               threexf % tuple(vects[0]),
                               threexf % tuple(vects[1]),
                               threexf % tuple(vects[2])])
    
    # Use system.symbols if set
    if symbols is None:
        if None not in system.symbols:
            symbols = system.symbols
    
    # Write symbols tags if they are given
    if symbols is not None:
        if not isinstance(symbols, (list, tuple)):
            symbols = [symbols]
        assert len(symbols) == system.natypes, 'length of symbols differs from number of atom types'
        poscar_string += '\n' + ' '.join(symbols)
    
    # Count how many atoms of each type
    atype = system.atoms.atype
    poscar_string += '\n' 
    uatype, counts = np.unique(atype, return_counts=True)
    
    for i in range(1, int(uatype.max()+1)):
        count = counts[uatype==i]
        if count == []:
            count = 0
        else:
            count = count[0]
        poscar_string += '%i ' % count
        
    # Check which coordinate style to use
    poscar_string += '\n' + coordstyle
    if coordstyle[0] in 'cCkK':
        scale = False
    else:
        scale = True
    
    # Write out positions
    pos = system.atoms_prop(key='pos', scale=scale)
    for a in range(1, system.natypes+1):
        for p in pos[atype==a]:
            poscar_string += '\n'+ threexf % tuple(p)
    
    # Save to the file-like object
    if hasattr(f, 'write'):
        f.write(poscar_string)
    
    # Save to the file name
    elif f is not None:
        with open(f, 'w') as fp:
            fp.write(poscar_string)
    
    # Return as a string
    else:
        return poscar_string


# coding: utf-8

# http://www.numpy.org/
import numpy as np

def pn_arctan_disregistry(xmax=None, xstep=None, xnum=None,
                          burgers=None, center=0.0,
                          halfwidth=1, normalize=True, shift=True):
    """
    Computes the classic Peierls-Nabarro arctan disregistry for an array of
    points x.
        
        δ(x) = b / π * arctan(x / ξ) + b / 2
    
    Parameters
    ----------
    xmax : float or None, optional
        Maximum value of x to use.  Minimum value is taken as -xmax.  At least
        2 of xmax, xstep, and xnum must be not None.  Default value is None.
    xstep : float or None, optional
        Step size to use between each x value.  At least 2 of xmax, xstep, and
        xnum must be not None.  Default value is None.
    xnum : int or None, optional
        Number of x values to use.  At least 2 of xmax, xstep, and xnum must
        be not None.  Default value is None.
    burgers : numpy.ndarray, optional
        The Burgers vector for the dislocation. Default value is [1, 0, 0].
    center : float
        The x coordinate to center the dislocation at. Default value is 0.0.
    halfwidth : float, optional
        The dislocation halfwidth to use. Default value is 1.
    normalize : bool, optional
        If True (default), the disregistry values will be scaled such that the
        two endpoints differ by exactly one Burgers vector.
    shift : bool, optional
        If True (default), the disregistry will range [0, 0, 0] to burgers.
        If False, the disregistry will range from -burgers to burgers.
        
    Returns
    -------
    x : numpy.ndarray
        The x-coordinates for the disregistry values.
    disregistry : numpy.ndarray
        The disregistry vector at each x-coordinate.
    """
    # Generate missing x parameters
    if xmax is None:
        if xstep is None or xnum is None:
            raise ValueError('At least two parameters must be given')
        xmax = xstep * (xnum - 1) / 2
    elif xstep is None:
        if xnum is None:
            raise ValueError('At least two parameters must be given')
        xstep = (2 * xmax) / (xnum - 1)
    elif xnum is None:
        xnum = ((2 * xmax) / xstep) + 1
    
    # Round xnum to int if needed
    if isinstance(xnum, float):
        if np.isclose(xnum, int(xnum + 0.5)):
            xnum = int(xnum + 0.5)
        else:
            raise ValueError('Invalid parameters: xnum or ((2 * xmax) / xstep) not an integer.')
    
    # Generate x and validate
    x, dx = np.linspace(-1*xmax, xmax, xnum, retstep=True)
    if not np.isclose(dx, xstep):
        raise ValueError('Incompatible parameters: xmax = xstep * (xnum - 1) / 2')
    
    if burgers is None:
        burgers = np.array([1.0, 0.0, 0.0])
    burgers = np.asarray(burgers)
    
    # δ(x) = b / π * arctan(x / ξ) + b / 2
    disregistry = np.outer(np.arctan((x - center) / halfwidth), burgers / np.pi) + burgers / 2
    
    if normalize is True:
        disregistry = disregistry - disregistry[0]
        disregistry = disregistry * np.linalg.norm(burgers) / np.linalg.norm(disregistry[-1])
    
    if shift is False:
        disregistry -= burgers / 2
    
    return x, disregistry


from . import gradient, integrator
from .BasePath import BasePath

from .ISMPath import ISMPath

def create_path(coord, energyfxn, style='ISM', gradientfxn='cdiff',
                gradientkwargs=None, integratorfxn='rk'):
    """
    Generic function call for initializing a Path object.
    
    Parameters
    ----------
    coord : array-like object
        The list of coordinates associated with the points along the path.
    energyfxn : function
        The function that evaluates the energy associated with the different
        point coordinates.
    style : str
        The relaxation style to associate with the path.  Determines the
        subclass to build.  Default value is "ISM" for the improved
        string method.
    gradientfxn : function, optional
        The function to use to estimate the gradient of the energy.  Default
        value of 'cdiff' will use atomman.mep.gradient.central_difference
    gradientkwargs : dict, optional
        The keyword arguments (i.e. settings) to use with the gradientfxn.
        Default is an empty dictionary, i.e. default settings of gradientfxn.
    integratorfxn : str or function, optional
        The function to use to integrate relaxation steps.  Default value of
        'rk' will use atomman.mep.integrator.rungekutta.
    """
    if style == 'ISM' or style == 'improved_string_method':
        return ISMPath(coord, energyfxn, gradientfxn=gradientfxn,
                       gradientkwargs=gradientkwargs,
                       integratorfxn=integratorfxn)
    else:
        raise ValueError('Unknown style')




# coding: utf-8
# atomman imports
from .aslist import aslist, iaslist
from .screen_input import screen_input
from .axes_check import axes_check
from .atomic_info import *
from .atomic_info import __all__ as atomic_info_all
from .compositionstr import compositionstr
from .duplicates_allclose import duplicates_allclose
from .vect_angle import vect_angle
from .uber_open_rmode import uber_open_rmode
from .indexstr import indexstr
from .filltemplate import filltemplate
from .crystalsystem import *
from .crystalsystem import __all__ as crystalsystem_all
from . import miller

__all__ = ['axes_check', 'compositionstr', 'duplicates_allclose', 'vect_angle',
           'uber_open_rmode', 'indexstr', 'filltemplate', 'miller', 'aslist',
           'iaslist', 'screen_input']
__all__.extend(atomic_info_all)
__all__.extend(crystalsystem_all)
__all__.sort()


# coding: utf-8
# Standard Python libraries
from io import BytesIO, open
import os

class uber_open_rmode():
    """
    Context manager for reading data from file-like objects, file names,
    and data strings in the same manner.
    """
    
    def __init__(self, data):
        """
        Initialize context manager.
        
        Parameters
        ----------
        data : file-like object or str
            Specifies what content to read.
        """
        self.data = data
        
    def __enter__(self):
        """Define different open actions."""
        def isfile(data):
            try:
                return os.path.isfile(self.data)
            except:
                return False
        
        # Any open file-like objects in 'r' mode will have a read attribute
        if hasattr(self.data, 'read'):
            self.open_file = self.data
            self.to_close = False
        
        # If data is path to a file, open the file
        elif isfile(self.data):
            self.open_file = open(self.data, 'rb')
            self.to_close = True
        
        # If data is not str, read using BytesIO
        elif not isinstance(self.data, str):
            self.open_file = BytesIO(self.data)
            self.to_close = True
            
        # If data is str, encode and read using BytesIO
        else:
            self.open_file = BytesIO(self.data.encode('utf-8'))
            self.to_close = True
        
        return self.open_file
    
    def __exit__(self, *args):
        """Close file if one was opened."""
        if self.to_close:
            self.open_file.close()